// naumann@stce.rwth-aachen.de

#include<iostream>
#include<cassert>
#include<cmath>

/// static vector type and arithmetic
template<int N, typename T>
class vector {
protected:
  T vals[N];
public:
  /// constructor initializes to zero
  vector() {
    for (size_t i=0;i<N;i++) vals[i]=T(0);
  }
  /// constructor initializes to val
  vector(const T& val) {
    for (size_t i=0;i<N;i++) vals[i]=T(val);
  }

  // norm
  T l2_norm() const {
    T n=0;
    for (int i=0;i<N;i++) n+=pow(vals[i],2);
    return sqrt(n);
  }

  /// read/write element access
  T& operator[](const size_t i) { 
    assert(i>=0&&i<N); 
    return vals[i]; 
  }

  /// read-only element access
  const T& operator[](const size_t i) const { 
    assert(i>=0&&i<N); 
    return vals[i]; 
  }

  /// element-wise addition
  vector operator+(const vector& v) const {
    vector w;
    for (size_t i=0;i<N;i++) w.vals[i]=vals[i]+v.vals[i];
    return w;
  }

  /// scaling
  vector operator*(const T& v) const {
    vector w;
    for (size_t i=0;i<N;i++) w[i]=v*vals[i];
    return w;
  }

  /// dot product
  T operator*(const vector& v) const {
    T w=0;
    for (size_t i=0;i<N;i++) w+=vals[i]*v.vals[i];
    return w;
  }

  // non-member friends
  template<int M, typename S>
  friend std::ostream& operator<<(std::ostream&, const vector<M,S>&);
  template<int M, typename S>
  friend std::istream& operator>>(std::istream&, vector<M,S>&);
};

/// output stream 
template<int N, typename T>
std::ostream& operator<<(std::ostream& s, const vector<N,T>& u) {
  s << "[ ";
  for (size_t i=0;i<N;i++) s << u.vals[i] << " ";
  s << "]";
  return s;
}

/// input stream 
template<int N, typename T>
std::istream& operator>>(std::istream& s, vector<N,T>& u) {
  for (size_t i=0;i<N;i++) s >> u.vals[i];
  return s;
}

/// static matrix type and matrix-matrix/vector arithmetic
template<int M, int N, typename T>
class matrix : public vector<M,vector<N,T>> {
using vector<M,vector<N,T>>::vals;
public:
  /// matrix-vector product
  vector<M,T> operator*(const vector<N,T>& x) const {
    vector<M,T> y;
    for (size_t i=0;i<M;i++) y[i]=vals[i]*x;
    return y;
  }

  /// matrix-matrix product
  template<int P>
  matrix<M,P,T> operator*(const matrix<N,P,T>& B) const {
    matrix<M,P,T> C;
    for (size_t i=0;i<M;i++) {
      C[i]=0;
      for (size_t j=0;j<N;j++) 
        C[i]=C[i]+B[j]*vals[i][j];
    }
    return C;
  }

  /// transpose
  matrix<N,M,T> transpose() const {
    matrix<N,M,T> C;
    for (size_t i=0;i<M;i++) {
      for (size_t j=0;j<N;j++) 
        C[j][i]=vals[i][j];
    }
    return C;
  }

};

int main() {
  const int n=2,m=3;
  vector<n,float> x;
  matrix<m,n,float> A;
  matrix<n,m,float> B;
  for (size_t i=0,k=1;i<n;i++) {
    x[i]=k++; 
    for (size_t j=0;j<m;j++) { A[j][i]=k++; B[i][j]=k++; }
  }
  std::cout << "A=" << A << std::endl;
  std::cout << "B=" << B << std::endl;
  std::cout << "x=" << x << std::endl;
  std::cout << "A*x=" << A*x << std::endl;
  std::cout << "A*B=" << A*B << std::endl;
  std::cout << "A.transpose()=" << A.transpose() << std::endl;
  std::cout << "A*A.transpose()=" << A*A.transpose() << std::endl;
  std::cout << "(A*B).transpose()=" << (A*B).transpose() << std::endl;
  // std::cout << "A*B.transpose()=" << A*B.transpose() << std::endl;
  std::cout << "(B*A)*A.transpose()=" << (B*A)*A.transpose() << std::endl;
  std::cout << "B*(A*A.transpose())=" << B*(A*A.transpose()) << std::endl;
  
  return 0;
}

 

